using System;
using System.Collections.Generic;
using System.Collections;
using System.Collections.Specialized;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Text.RegularExpressions;
using System.Windows.Forms;
using SpringPCSCAPI;
using System.Runtime.InteropServices;

namespace CardTool
{
  public partial class RWCardFrm : Form
  {

#region constants

    DataTable CardDataTable;
    DataTable CardAsciiTable;
    int nb_blocks=0;
    int sz_blocks=0;
    int MAX_DATA_OUT = 254;
    int COL_WIDTH = 25;
    int ROW_HEIGHT = 23;
    int COL_HEADER_HEIGHT = 8;
    int le_max = 254;
    byte[] cardContent;
    ReaderStatusFrm parentFrm;
    Button[] WritingButtons;
    NameValueCollection keys;
    KeysFrm KeysDlg;
    string type = "";
    string InfoText = "";
    //string LogText = "";
    string lastApduIn = "";
    string lastApduOut = "";
    int cpt_read_done;
    int CurrentLine = -1;
    //int CurrentLineHex = -1;
    //int CurrentLineASCII = -1;
    int CurrentLineChanged = -1;
    /// <summary>
    /// The height in pixels of a line.
    /// </summary>
    private const float ciLineHeight = 14.32F;
    private const int minAscii = 24;
    private const int minHeight = 240;
    private const int ciHexWidth = 21;
    private const int ciAsciiWidth = 9;
    bool checkLine = false;

    //mousewheel
    private int fontSize = 20;
    private System.Drawing.Drawing2D.GraphicsPath mousePath;


    /* we use color codes to determine writing access for one block :
     * - green : the block is writable this is the default configuration
     * - orange : the block has to be written with care (OTP), 
     *     ask for prompt before writing 
     * - red : cannot write block
     */
    string[] writable;

    System.Threading.Thread ReadThread = null; /* Read card */
    public ThreadFinishedDelegate ThReadDeleg;
    public delegate void ThreadFinishedDelegate();
#endregion

    public RWCardFrm(string type_card, ReaderStatusFrm parent)
    {
      InitializeComponent();

      /* mousewheel */
      richTextBoxASCII.MouseWheel += new MouseEventHandler(richTextBoxASCII_MouseWheel);
      richTextBoxHex.MouseWheel += new MouseEventHandler(richTextBoxHex_MouseWheel);
      richTextBoxBlock.MouseWheel += new MouseEventHandler(richTextBoxBlock_MouseWheel);
      richTextBoxInfo.MouseWheel += new MouseEventHandler(richTextBoxInfo_MouseWheel);


      keys = new NameValueCollection();
      keys.Add("00", "");
      keys.Add("01", "");
      keys.Add("02", "");
      keys.Add("03", "");
      keys.Add("10", "");
      keys.Add("11", "");
      keys.Add("12", "");
      keys.Add("13", "");
      //ReadFullCard(type_card);
      parentFrm = parent;
      type = type_card;
      cpt_read_done = 0;

      ThStatusDeleg = new StatusThreadFinishedDelegate(onStatusDone);

      // mifare 1K & 4K 
     /* if ((type == "0001") || (type == "0002"))
        SetKeysBut.Visible = true;*/

      ReadIniFiles();
      cardContent = new byte[sz_blocks * nb_blocks];
      writable = new string[nb_blocks];
      // initialize card content to 0x00
      for (int i = 0; i < sz_blocks * nb_blocks; i++)
        cardContent[i] = 0x00;
      ThReadDeleg = new ThreadFinishedDelegate(FullLoadDataTable);
      // initialize writing access table for each block
      for (int i = 0; i < nb_blocks; i++)
        writable[i] = "green";
      WritableBlocks();
      StartReading();

    }

#region mouse wheel

    void richTextBoxInfo_MouseWheel(object sender, MouseEventArgs e)
    {
      panelMain.Focus();
    }

    void richTextBoxBlock_MouseWheel(object sender, MouseEventArgs e)
    {
      panelMain.Focus();
    }

    void richTextBoxHex_MouseWheel(object sender, MouseEventArgs e)
    {
      panelMain.Focus();
    }

    void richTextBoxASCII_MouseWheel(object sender, MouseEventArgs e)
    {
      panelMain.Focus();
    }
#endregion

#region start up functions
    public void ReadIniFiles()
    { 
     // search ini file to find card structure according to type
      nb_blocks = Convert.ToInt32(IniReadValue(type, "NB"));
      sz_blocks = Convert.ToInt32(IniReadValue(type, "SZ"));
      cardTypeTb.Text = IniReadValue(type, "Name");
      this.Text += " : " + cardTypeTb.Text;
      
      le_max = Convert.ToInt32(IniReadValue(type,"LE"));
    }

    public void WritableBlocks()
    {
      char[] delimiterMinMax = {'-'};
      char[] delimiterList = {','};
      // are there orange type of blocks?
      string orangeBlocks = IniReadValue(type, "ORANGE");
      // if value is NO, no orange block
      if (orangeBlocks != "NO")
      {
        // if value is x-y , x to y blocks are orange
        if (orangeBlocks.Contains("-"))
        {
          string[] MinMax = orangeBlocks.Split(delimiterMinMax);
          int min = Convert.ToInt32(MinMax[0]);
          int max = Convert.ToInt32(MinMax[1]);
          for (int i = min; i < max+1; i++)
            writable[i] = "orange";
        }
        else
        {
          if (orangeBlocks.Contains(","))
          {
            // if value is x,y,...,z , x and y and ... and z are orange
            string[] listOrange = orangeBlocks.Split(delimiterList);
            for (int i = 0; i < listOrange.Length; i++)
              writable[Convert.ToInt32(listOrange[i])] = "orange";
          }
          else
            // if value is x, only x is orange
            writable[Convert.ToInt32(orangeBlocks)] = "orange";
          
        }
      }
      // are there red type of blocks?
      string redBlocks = IniReadValue(type, "RED");
      // if value is NO, no red block
      if (redBlocks != "NO")
      {
        // if value is x-y , x to y blocks are red
        if (redBlocks.Contains("-"))
        {
          string[] MinMax = redBlocks.Split(delimiterMinMax);
          int min = Convert.ToInt32(MinMax[0]);
          int max = Convert.ToInt32(MinMax[1]);
          for (int i = min; i < max; i++)
            writable[i] = "red";
        }
        else
        {
          if (redBlocks.Contains(","))
          {
            // if value is x,y,...,z , x and y and ... and z are red
            
            string[] listRed = redBlocks.Split(delimiterList);
            for (int i = 0; i < listRed.Length; i++)
            {
              writable[Convert.ToInt32(listRed[i])] = "red";
            }
          }
          else
            // if value is x, only x is red
            writable[Convert.ToInt32(redBlocks)] = "red";

        }
      }
    
    }

    public void ReadFullCard()
    { 
     
      
      // read card and upload content to DataGrid 
      if ((type != "0001") && (type != "0002") && (type != "0200") && (type != "0400"))
        ReadCard();
      else
        ReadMifareCard(type);
      //FullLoadDataTable();
      try
      {
        Invoke(ThReadDeleg);
      }
      catch (Exception e) { }
    }
#endregion

#region read ini file

    [DllImport("kernel32")]
    private static extern int GetPrivateProfileString(string section,
        string key, string def, StringBuilder retVal, int size, string filePath);
    public string IniReadValue(string Section, string Key)
    /* get initialization data from ini file */
    {
      StringBuilder temp = new StringBuilder(500);
      String CurrentDir = Application.StartupPath;
      int i = GetPrivateProfileString(Section, Key, "", temp, 500, CurrentDir + "/type_card.ini");
      return temp.ToString();
    }
#endregion

#region read card

    public void ReadCard()
    {
      if ((nb_blocks > 0) && (sz_blocks > 0) &&
        (sz_blocks * nb_blocks < MAX_DATA_OUT)
        && (le_max >= sz_blocks * nb_blocks))
      // read full card with just one APDU
      {
        //size blocks * nb blocks < 254 
        // therefore a simple cast to byte can be performed
        byte LE = (byte)(sz_blocks * nb_blocks);
        string apdu_read_full = "FF B0 00 00 " +
          String.Format("{0:X02}", LE);
        APDUCmd CmdRead = new APDUCmd(apdu_read_full);
        APDURsp RspRead = ConnectAndTransmit(CmdRead);
        
        //if we received expected length
        if (RspRead.APDUArray.Length == nb_blocks * sz_blocks+2)
        {
          cardContent = RspRead.DataOut;
          cpt_read_done = nb_blocks * sz_blocks;
        }
      }
      else
      {
        if (le_max < sz_blocks * nb_blocks)
        // length expected for read command is limited
        {
          ReadCard_Sequential();
        }
      }
    }

    public void ReadCard_Sequential()
    {
      int cpt_content = 0,i,j,rest;

      int from_offset = 0;
      SPCSC.ConnectCard();
      for(i=0;i<(sz_blocks*nb_blocks/le_max);i++)
      {
        
        APDUCmd CmdRead = new APDUCmd(0xFF, 0xB0,
          (byte)((from_offset & 0xFF00) >> 8),
          (byte)(from_offset & 0x00FF),
          (byte)le_max);
        APDURsp RspRead = Transmit(CmdRead);
        if ((RspRead.APDUArray.Length > 2) && (RspRead.SW1==0x90) && (RspRead.SW2==0x00))
        {
          for (j = 0; j < RspRead.DataOut.Length; j++)
            cardContent[cpt_content + j] = RspRead.DataOut[j];
          cpt_content += le_max;
          cpt_read_done = cpt_content;
          from_offset++;
        }
        else
        {
          break;
        }
      
      }
      rest = (sz_blocks * nb_blocks) % le_max;
      if (rest > 0)
      {
        APDUCmd CmdRead = new APDUCmd(0xFF, 0xB0,
          (byte)(((cpt_content / sz_blocks) & 0xFF00) >> 8),
          (byte)((cpt_content/sz_blocks) & 0x00FF),
          (byte)rest);
        APDURsp RspRead = Transmit(CmdRead);
        if ((RspRead.APDUArray.Length > 2) && (RspRead.SW1==0x90) && (RspRead.SW2==0x00))
        {
          for (j = 0; j < RspRead.DataOut.Length; j++)
            cardContent[cpt_content + j] = RspRead.DataOut[j];
          cpt_read_done = cpt_content + j - 1;
        }

      }
      SPCSC.Disconnect();

    }

    public void ReadMifareCard(string type)
    {
      byte sector;
      APDURsp resp;
      int cpt_content = cpt_read_done,i;
      
      int nb_sectors = 16;
      int from_sector=0;
      
      if (cpt_read_done > 0)
      {
        if (cpt_read_done <= 2048)
          from_sector = cpt_read_done / (16 * 4);
        else
          from_sector = ((cpt_read_done - 2048) / 256) + 32;
      
      }
      //mifare 4k
      if ( (type == "0002") ||(type=="0200"))
        nb_sectors = 40;

      for (sector = (byte)from_sector; sector < nb_sectors; sector++)
      {

        /*byte block = (byte)((int)sector * 4);
        byte length = 0x30;
        resp = new APDURsp();
        APDUCmd ReadMifApdu = new APDUCmd(0xFF, 0xF3, 0x00, block, length);
        bool retTransmit = SPCSC.Transmit(ReadMifApdu, ref resp);*/
        byte block = (byte)((int)sector * 4);
        byte length = 0x30;

        // mifare 4k
        if ((int)sector > 31)
        {
          block = (byte)(128 + 16 * ((int)sector - 32));
          length = 0xF0;
        }
        SPCSC.ConnectCard();
        APDUCmd ReadMifApdu = new APDUCmd(0xFF, 0xF3, 0x00, block, length);
        resp = Transmit(ReadMifApdu);
        // if data received 
        if ((resp.APDUArray.Length >2) && (resp.SW1 ==0x90) && (resp.SW2==0x00))
        {
          for (i = 0; i < resp.DataOut.Length; i++)
            cardContent[cpt_content + i] = resp.DataOut[i];
          cpt_content += resp.DataOut.Length;

          cpt_content += sz_blocks;
          cpt_read_done = cpt_content;
          SPCSC.Disconnect();
          InfoText += "\n\nKey to read sector " + sector.ToString() + " :\n";
          getLastAuthKey();
          
        }
        /*else
        {
          //if ((resp.SW1 == 0x69) && (resp.SW2 == 0x82))
          //  AuthAndReadMif(sector);
          break;
        }*/
        
      }



     
    }

    public APDURsp AuthAndReadMif(byte sector)
    {
      // in that case the authentication is done with a key
      // set in non volatile memory of the reader (via Load Keys form)
      APDURsp resp = new APDURsp();
      // for each key set by user on volatile memory( keys[])
   
      
      for (byte i = 0x00; i < 0x04; i++)
      {
        // checking if key is set
        if (keys[byteToString(i)] != "")
        {
          MessageBox.Show(byteToString(i) + ": Trying with key :" + keys[byteToString(i)]);
          // try to read with that key
          SPCSC.ConnectCard();
          resp = SPCSC.ReadMif(sector, i);
          // if read succeeded, return the response and display
          // key index/key value that was valid for authentication on this section
          MessageBox.Show("RESp : " + resp.APDUStr);
          if ((resp.SW1 == 0x90) && (resp.SW2 == 0x00))
          {
            lastApduOut = resp.APDUStr;
            InfoText += "Reading sector " + sector.ToString() +
              " with key " + i.ToString() + " (" + keys[byteToString(i)] + ") ";
            break;

          }
        }

      }
      if ((resp.SW1 != 0x90) || (resp.SW2 != 0x00))
      { 
        // retry with B keys
        for (byte i = 0x10; i < 0x14; i++)
        {
          // checking if key is set
          if (keys[byteToAscii(i)] != "")
          {
            // try to read with that key
            SPCSC.ConnectCard();
            resp = SPCSC.ReadMif(sector, i);
            // if read succeeded, return the response and display
            // key index/key value that was valid for authentication on this section
            if ((resp.SW1 == 0x90) && (resp.SW2 == 0x00))
            {
              lastApduOut = resp.APDUStr;
              InfoText += "Reading sector " + sector.ToString() +
                " with key " + i.ToString() + " (" + keys[i] + ") ";
              break;

            }
          }

        }
      }


      return resp;
    
    }

    /*public APDURsp ReadMifareSector(byte sector)
    { 
      // First thing to do : Mifare classic read
      // in that case the authentication is done with a key
      // set in non volatile memory of the reader


      // if read failed, 

      // for each key set by user on volatile memory, keys[]
      // try to read with that key
      // checking if key is set
      // from 00 to 03
      // if read succeeded, return the response and display
      // key index/key value that was valid for authentication on this section

    }*/

#endregion

#region load data richtextbox
    public void FullLoadDataTable()
    {
      int i,j,cpt=0;
      checkLine = false;
      NbBlocksTb.Text = nb_blocks.ToString();
      SizeBlocksTb.Text = sz_blocks.ToString();
      
      int widthAscii = ciAsciiWidth * sz_blocks;
      if(widthAscii <minAscii)
        widthAscii = minAscii;
      int height = (int)(ciLineHeight * nb_blocks);
      if(height<minHeight)
        height = minHeight;
     

      CardDataTable = new DataTable();
      CardAsciiTable = new DataTable();
      
      
      // adapt DataGridView to card type 
    /*  cardContentDg.Size = new Size(COL_WIDTH * sz_blocks,
        ROW_HEIGHT * nb_blocks + COL_HEADER_HEIGHT);
      CardAsciiDg.Size = new Size(COL_WIDTH * sz_blocks,
        ROW_HEIGHT * nb_blocks + COL_HEADER_HEIGHT);
      cardContentDg.Location = new Point(CardAsciiDg.Location.X +
        CardAsciiDg.Size.Width + 20, cardContentDg.Location.Y);*/
      //CardAsciiDg.Size = new Size(38 * sz_blocks, 24 * nb_blocks);

      richTextBoxASCII.Size = new Size(widthAscii,
        height);
      richTextBoxHex.Size = new Size(ciHexWidth*sz_blocks,
       height);
      richTextBoxBlock.Size = new Size(richTextBoxBlock.Size.Width,
        height);
      richTextBoxInfo.Size = new Size(richTextBoxInfo.Size.Width,
        height);
      richTextBoxASCII.Location = new Point(
        richTextBoxHex.Location.X + richTextBoxHex.Size.Width + 5,
        richTextBoxASCII.Location.Y);
      richTextBoxInfo.Location = new Point(
        richTextBoxASCII.Location.X + richTextBoxASCII.Size.Width + 5,
        richTextBoxInfo.Location.Y);
      /*if (nb_blocks > 20)
      {
        this.Size = new Size(1100, 750);
      }*/
     
      //create columns according to blocks size
     /* DataColumn col, colAscii;
      for (i = 0; i < sz_blocks; i++)
      {
        col = new DataColumn();
        colAscii = new DataColumn();
        col.DataType = System.Type.GetType("System.String");
        colAscii.DataType = System.Type.GetType("System.String");
        col.ColumnName = i.ToString();
        colAscii.ColumnName =i.ToString();
        CardDataTable.Columns.Add(col);
        CardAsciiTable.Columns.Add(colAscii);
      }

      //creating and loading rows according to the number of blocks
      DataRow Page;
      DataRow PageAscii;*/
      richTextBoxBlock.Clear();
      richTextBoxASCII.Clear();
      richTextBoxHex.Clear();
      for (i = 0; i < nb_blocks; i++)
      {
        //Page = CardDataTable.NewRow();
       // PageAscii = CardAsciiTable.NewRow();
        richTextBoxBlock.AppendText(i.ToString());
        for (j = 0; j < sz_blocks; j++)
        {
         // Page[ j.ToString()] = 
         //  String.Format("{0:X02}", cardContent[cpt]);
          richTextBoxHex.AppendText(String.Format("{0:X02}", cardContent[cpt]));
          if(j<sz_blocks-1)
            richTextBoxHex.AppendText(" ");
         //PageAscii[j.ToString()] =
          //  byteToAscii(cardContent[cpt]);
          richTextBoxASCII.AppendText(byteToAscii(cardContent[cpt]));
          
          cpt ++;
        }
       // CardAsciiTable.Rows.Add(PageAscii);
       // CardDataTable.Rows.Add(Page);
        if (i < nb_blocks - 1)
        {
          richTextBoxBlock.AppendText("\n");
          richTextBoxASCII.AppendText("\n");
          richTextBoxHex.AppendText("\n");
        }
      }
     // DataGridTableStyle ts = new DataGridTableStyle();
     // cardContentDg.TableStyles.Add(ts);
     // ts.MappingName = CardDataTable.TableName;
     // cardContentDg.DataSource = new DataView(CardDataTable);
     // CardAsciiDg.DataSource = new DataView(CardAsciiTable);
      //cardContentDg.TableStyles[0].GridColumnStyles[0].Width = 100;

      //CreateWriteButtons();
      //textBox1.Text = InfoText;
      richTextBoxInfo.Text = InfoText;


      ApduRecvTb.Text = lastApduOut;
      ApduSentTb.Text = lastApduIn;
      /*if (InfoText != "")
      {
        butRead.Text = "Start Reading";
        butPause.Visible = false;
        ButStart.Visible = true;
        ButStart.Enabled = true;
      }
      if (cpt_read_done == sz_blocks * nb_blocks)
      {
        butRead.Visible = false;
        butPause.Visible = false;
        ButStart.Visible = true;
        ButStart.Enabled = true;
        SetKeysBut.Enabled = true;
      }*/
      //butPause.Visible = false;
      //ButStart.Visible = true;
      //ButStart.Enabled = true;
      //SetKeysBut.Enabled = true;
     

      checkLine = true;
      panelMain.Focus();
    }

    private void getLastAuthKey()
    {
     
      // only if card is Mifare 1K or 4K
      if ((type == "0001") || (type == "0002") || (type == "0200") || (type == "0400"))
      {
       
        SPCSC.directConnect();
        APDURsp key = new APDURsp();
        byte[] getKey = {0x58,0x33};
        int ret = SPCSC.Control(getKey,ref key);
        //MessageBox.Show(ret.ToString()+ SPCSC.codeToStr((uint) ret));
        //textBox1.Text = key.APDUStr;
        SPCSC.Disconnect();
        // if command succeeded
       
        if ((key.APDUArray[0] == 0x00) && (key.APDUArray.Length == 2))
          decodeKey(key.APDUArray[1]);
      }
    }

    private void decodeKey(byte key)
    {
      
      byte RAM = 0x80;
      byte EEPROM = 0x40;
      byte B_KEY = 0x20;
      byte index = (byte) ((key) & 0x0F);

      if ((key & RAM) !=0)
        InfoText += "key in RAM ";
      if ((key & EEPROM) != 0)
        InfoText += "key in EEPROM ";
      if ((key & B_KEY) != 0)
        InfoText += "- B ";
      else
        InfoText += "- A ";
      InfoText += "- " + byteToString(index)+"\n";
    }




    public string byteToAscii(byte value)
    {
      if (value < 32 || value > 127)
        return(".");
      else
        return (System.Convert.ToChar(value).ToString());
    }

    public string byteToString(byte value)
    {
      if (value < 32 || value > 127)
        return (String.Format("{0:X02}", value));
      else
        return (System.Convert.ToChar(value).ToString());
    }

    private void WritingButtonW_Click(object sender, EventArgs e)
    {
      if( MessageBox.Show("It is not recommended that you write data on this block.\n" +
        "Make sure that you have fully understood card documentation before writing.\n" +
        "Continue anyway?","Warning",MessageBoxButtons.OKCancel) != DialogResult.Cancel )
      {
        for (int i = 0; i < nb_blocks; i++) 
        {
          if (sender.Equals(WritingButtons[i]))
            WriteBlock(i);
        }

      }
    }

    private void WritingButton_Click(object sender, EventArgs e)
    {
      for (int i = 0; i < nb_blocks; i++) 
      {
        if (sender.Equals(WritingButtons[i]))
          WriteBlock(i);
      }
    }

    public void WriteBlock(int block)
    { 
      // APDU is FF D6 P1 P2 LC + size of a block
      byte[] NewData = new byte[sz_blocks];
      APDUCmd WriteApdu;
       
      // retreive data from datagrid
      for (int i = 0; i < sz_blocks; i++)
        NewData[i] = 
          System.Convert.ToByte(int.Parse(
          cardContentDg.Rows[block].Cells[i].Value.ToString(),
          System.Globalization.NumberStyles.HexNumber));
      // Mifare requires specific command : F4 is Mifare Classic Write
      if ((type != "0001") && (type != "0002") && (type != "0200") && (type != "0400"))
        WriteApdu = new APDUCmd(0xFF,0xD6,(byte)((block & 0xFF00) >> 8),
        (byte)((block) & 0x00FF),NewData);
      else
        WriteApdu = new APDUCmd(0xFF, 0xF4, (byte)((block & 0xFF00) >> 8),
        (byte)((block) & 0x00FF), NewData);
      MessageBox.Show("Writing data to card with APDU :\n"+WriteApdu.APDUStr);
      ConnectAndTransmit(WriteApdu);
      ApduSentTb.Text = lastApduIn;
      ApduRecvTb.Text = lastApduOut;
      CurrentLine = -1;
      // change CardContent Array & CardContentAscii
      for (int i = 0; i < sz_blocks; i++)
      {
        cardContent[block * sz_blocks + i] = System.Convert.ToByte(int.Parse(
          cardContentDg.Rows[block].Cells[i].Value.ToString(),
          System.Globalization.NumberStyles.HexNumber));
        CardAsciiDg.Rows[block].Cells[i].Value =
          byteToAscii(cardContent[block * sz_blocks + i]);
      }
    }

    public void WriteLine(int line)
    {
      // APDU is FF D6 P1 P2 LC + size of a block
      byte[] NewData = new byte[sz_blocks];
      APDUCmd WriteApdu;
      char[] NewLine = { '\n' };
      string[] Line = GetLine(richTextBoxHex, line).Split(NewLine);
      string TextHex = Line[0]; 
      string TextAscii = "";
      char[] delimiter = { ' ' };
      string[] DataHex = TextHex.Split(delimiter);

      // retreive data from datagrid
      for (int i = 0; i < sz_blocks; i++)
        NewData[i] =
          System.Convert.ToByte(int.Parse(
          DataHex[i],
          System.Globalization.NumberStyles.HexNumber));
      // Mifare requires specific command : F4 is Mifare Classic Write
      if ((type != "0001") && (type != "0002") && (type!="0200") && (type!="0400"))
        WriteApdu = new APDUCmd(0xFF, 0xD6, (byte)((line & 0xFF00) >> 8),
        (byte)((line) & 0x00FF), NewData);
      else
        WriteApdu = new APDUCmd(0xFF, 0xF4, (byte)((line & 0xFF00) >> 8),
        (byte)((line) & 0x00FF), NewData);
      //MessageBox.Show("Writing data to card with APDU :\n" + WriteApdu.APDUStr);
      ConnectAndTransmit(WriteApdu);
      richTextBoxInfo.Text += "Sending " + lastApduIn + "\n";
      richTextBoxInfo.Text += "Received " + lastApduOut + "\n";
      // scroll bar
      // Put the cursor at the very end of the text
      richTextBoxInfo.Select((richTextBoxInfo.Text.Length), 0);

      // Scroll
      richTextBoxInfo.ScrollToCaret();
      /*
      ApduSentTb.Text = lastApduIn;
      ApduRecvTb.Text = lastApduOut;*/
      CurrentLine = -1;
      // change CardContent Array & CardContentAscii
      for (int i = 0; i < sz_blocks; i++)
      {
        cardContent[line * sz_blocks + i] = NewData[i]; /*System.Convert.ToByte(int.Parse(
          cardContentDg.Rows[line].Cells[i].Value.ToString(),
          System.Globalization.NumberStyles.HexNumber));
        //CardAsciiDg.Rows[line].Cells[i].Value =
         // byteToAscii(cardContent[line * sz_blocks + i]);*/
      }
    }

    public APDURsp ConnectAndTransmit(APDUCmd ApduIn)
    {
     
      SPCSC.ConnectCard();
      APDURsp Resp = Transmit(ApduIn);
      SPCSC.Disconnect();
      

      return Resp;
    }

    public APDURsp Transmit(APDUCmd ApduIn)
    {
      APDURsp Resp = new APDURsp();
      SPCSC.Transmit(ApduIn, ref Resp);
     // InfoText += "Sending " + ApduIn.APDUStr + "\n"; 
      lastApduIn = ApduIn.APDUStr;
      lastApduOut = Resp.APDUStr + " - " + SPCSC.StatusWordsToStr(Resp.SW1, Resp.SW2);
      //InfoText += "Received " + Resp.APDUStr + " - " + SPCSC.StatusWordsToStr(Resp.SW1, Resp.SW2) + "\n";
      if ((Resp.SW1 != 0x90) || (Resp.SW2 != 0x00))
      {
        string ErrorText = "Error while sending " + ApduIn.APDUStr
            + " :\n " + Resp.APDUStr + " - " + SPCSC.StatusWordsToStr(Resp.SW1, Resp.SW2);
        InfoText += ErrorText + "\n";
        //MessageBox.Show(ErrorText);
      }
      return Resp;
    }

#endregion

#region conversion from str to byte array and vice versa

    private byte[] createByteArrayFromStr(string StrIn)
    {
      string Str = StrIn.Replace("0x", "").Replace(" ","");

      if (!Str.Contains(" "))
      {
        string nStr = "";
        for (int cpt = 0; cpt < Str.Length; cpt = cpt + 2)
        {
          nStr = nStr + Str.Substring(cpt, 2) + " ";
        }
        Str = nStr;
      }

      // command stored into byte array
      char[] delimiterByte = new char[1];
      delimiterByte[0] = ' ';
      string properStr = Str;
      // removing last " "
      if (Str.EndsWith(" "))
      {
        properStr = Str.Substring(0, Str.Length - 1);
      }

      string[] Strbuff = properStr.Split(delimiterByte);
      byte[] StrBuffer = new byte[Strbuff.Length];
      for (int i = 0; i < Strbuff.Length; i++)
      {
        // converting byte string to byte  
        StrBuffer[i] = System.Convert.ToByte(int.Parse(Strbuff[i],
          System.Globalization.NumberStyles.HexNumber));
      }
      return StrBuffer;
    }

    private string createStrFromByteArray(byte[] byteArray)
    {
      string Str = "";
      if (byteArray.Length <= 0)
        return Str;
      for (int cpt = 0; cpt < byteArray.Length - 1; cpt++)
        Str += String.Format("{0:X02}", byteArray[cpt]) + " ";
      Str += String.Format("{0:X02}", byteArray[byteArray.Length - 1]);
      return Str;
    }
#endregion

#region set keys

    private void SetKeysBut_Click(object sender, EventArgs e)
    {
      KeysDlg = new KeysFrm(this, keys);
      KeysDlg.Show();   
    }

    public void setKeys(NameValueCollection keys_in)
    { 
      // input paramater is a collection of keys
      // a key is a 6 bytes array represented by a string
      keys = keys_in;
      SPCSC.ConnectCard();
      foreach (string s in keys_in.AllKeys)
      // s is Key index and keys_in[s] is key value
      {
        if (keys_in[s] != "")
        {
          SPCSC.LoadKey(createByteArrayFromStr(keys_in[s]),
            System.Convert.ToByte(int.Parse(s,
          System.Globalization.NumberStyles.HexNumber)));
          richTextBoxInfo.Text += "Loading key ["+s+"] "+
          keys_in[s]+"\n";
          // scroll bar
          // Put the cursor at the very end of the text
          richTextBoxInfo.Select((richTextBoxInfo.Text.Length), 0);

          // Scroll
          richTextBoxInfo.ScrollToCaret();
          
        }
        /*SPCSC.SetKey(createByteArrayFromStr(keys_in[s]),
          System.Convert.ToByte(int.Parse(s,
          System.Globalization.NumberStyles.HexNumber)));*/
        
      }
      SPCSC.Disconnect();
     // ButStart.Enabled = true;
     // ButStart.Visible = true;
      //APDURsp a = SPCSC.LoadKeys();
      //MessageBox.Show(a.APDUStr);
    }

#endregion

#region former code (with now disabled controls)

    public void CreateWriteButtons()
    {
      /* Creation of the writing buttons */
      WritingButtons = new Button[nb_blocks];
      for (int i = 0; i < nb_blocks; i++)
      {
        if (writable[i] != "red")
        {
          WritingButtons[i] = new Button();
          // WritingButtons[i].Enabled = false;
          WritingButtons[i].Location = new System.Drawing.Point(
            (cardContentDg.Location.X + COL_WIDTH * sz_blocks + 15)
            , cardContentDg.Location.Y + (i + 1) * 22);
          //, 38 + (i + 2) * 22);
          WritingButtons[i].Name = "button" + i.ToString();
          WritingButtons[i].Size = new System.Drawing.Size(41, 22);
          WritingButtons[i].Image = CardTool.Properties.Resources.rec_icon;
          WritingButtons[i].UseVisualStyleBackColor = true;
          if (writable[i] == "orange")
            WritingButtons[i].Click += new System.EventHandler(this.WritingButtonW_Click);
          else
            WritingButtons[i].Click += new System.EventHandler(this.WritingButton_Click);
          this.Controls.Add(WritingButtons[i]);
        }
      }

    }

    private void butRead_Click(object sender, EventArgs e)
    {
      MessageBox.Show("butRead is Disabled!");
      /* butRead.Enabled = false;
      ReadFullCard(type);
      butRead.Visible = false;*/
      if (butRead.Text == "Start Reading")
      {
        butRead.Text = "Stop Reading";
        ReadThread = new System.Threading.Thread(new
        System.Threading.ThreadStart(ReadFullCard));
        ReadThread.Start();
      }
      else
        if (butRead.Text == "Stop Reading")
        {
          butRead.Text = "Resume Reading";
          if (ReadThread != null)
          {
            try
            {
              ReadThread.Suspend();
              //MessageBox.Show("Done reading : " + cpt_read_done);
              SPCSC.Disconnect();
              FullLoadDataTable();
            }
            catch (Exception x)
            {
              MessageBox.Show(x.Message);
            }
          }
        }
        else
        {
          if (ReadThread != null)
          {
            try
            {
              ReadFullCard();//ReadThread.Resume();
              FullLoadDataTable();
            }
            catch (Exception x)
            {
              MessageBox.Show(x.Message);
            }
          }
        
        }
    }

    private void cardContentDg_CellEndEdit(object sender, DataGridViewCellEventArgs e)
    {
      /* Checking if value entered is hexadecimal */
      String newValue = System.Convert.ToString(
        cardContentDg.Rows[e.RowIndex].Cells[e.ColumnIndex].Value);
      Regex regHex = new Regex("^[0-9a-fA-f]{1,2}$");
      if (!regHex.IsMatch(newValue))
      {
        MessageBox.Show(
          "Only hexadecimal values between 0 and FF allowed.");
        cardContentDg.Rows[e.RowIndex].Cells[e.ColumnIndex].Value =
          String.Format("{0:X02}", cardContent[e.RowIndex*sz_blocks+ e.ColumnIndex]);
      }
      else
      {
        if (cardContentDg.Rows[e.RowIndex].Cells[e.ColumnIndex].Value.ToString() !=
          String.Format("{0:X02}", cardContent[e.RowIndex * sz_blocks + e.ColumnIndex]) ) 
          CurrentLine = e.RowIndex;
      }
    }

    private void cardContentDg_CellBeginEdit(object sender, DataGridViewCellCancelEventArgs e)
    {
      if (writable[e.RowIndex] == "red")
      {
        MessageBox.Show("Cells from this line cannot be changed");
        e.Cancel = true;
      }
      if (CurrentLine != -1 && CurrentLine != e.RowIndex)
      {
        String message =
          "You haven't recorded block " +
          CurrentLine.ToString() + "\n" + "Would you like to write it to card ?";

        if (MessageBox.Show(message, "Record", MessageBoxButtons.YesNo)
          == DialogResult.No)
        {
          /* Rollback Changes made to CurrentLine*/
          for (int i = 0; i < sz_blocks; i++)
            cardContentDg.Rows[CurrentLine].Cells[i].Value =
               String.Format("{0:X02}",cardContent[CurrentLine * sz_blocks + i]);
        }
        else
        {
          if (writable[e.RowIndex] == "orange")
          { 
            /* Ask before Writing */
            if (MessageBox.Show("It is not recommended that you write data on this block.\n" +
        "Make sure that you have fully understood card documentation before writing.\n" +
        "Continue anyway?", "Warning", MessageBoxButtons.OKCancel) != DialogResult.Cancel)
                 WriteBlock(CurrentLine);
          }
          else
          {
            WriteBlock(CurrentLine);
          }
        }
        CurrentLine = -1;
      }
    }
#endregion

#region buttons

    private void ButStart_Click(object sender, EventArgs e)
    {
      StartReading();
    }

    private void StartReading()
    {
      //ButStart.Enabled = false;
      //SetKeysBut.Enabled = false;
      if (ReadThread != null)
      {
        try
        {
          ReadFullCard();//ReadThread.Resume();
          FullLoadDataTable();
        }
        catch (Exception x)
        {
          MessageBox.Show(x.Message);
        }
      }
      else
      {

        //butPause.Visible = true;
       //ButStart.Visible = false;
        ReadThread = new System.Threading.Thread(new
        System.Threading.ThreadStart(ReadFullCard));
        ReadThread.Start();
      }
    }

    private void butPause_Click(object sender, EventArgs e)
    {
      //ButStart.Visible = true;
      //ButStart.Enabled = true;
      //SetKeysBut.Enabled = true;
      //butPause.Visible = false;
      if (ReadThread != null)
      {
        try
        {
          ReadThread.Suspend();
          //MessageBox.Show("Done reading : " + cpt_read_done);
          SPCSC.Disconnect();
          FullLoadDataTable();
        }
        catch (Exception x)
        {
          MessageBox.Show(x.Message);
        }
      }
    }

#endregion

#region richtextbox management
    private void richTextBoxHex_TextChanged(object sender, EventArgs e)
    {

      int cursor = richTextBoxHex.SelectionStart;
      if (checkLine)
      {

        
        if ((CurrentLineChanged == -1)
          || (CurrentLineChanged == richTextBoxHex.GetLineFromCharIndex(
            richTextBoxHex.SelectionStart)))
        {
          CurrentLineChanged = richTextBoxHex.GetLineFromCharIndex(
              richTextBoxHex.SelectionStart);
          toolStripLabel.Text = "You are modifying line " + CurrentLineChanged.ToString() ;
          if (writable[
            richTextBoxHex.GetLineFromCharIndex(
            richTextBoxHex.SelectionStart)] == "red")
          {
            MessageBox.Show("Data from this line cannot be changed.");

            RollBackLineHex(richTextBoxHex, CurrentLineChanged);
            CurrentLineChanged = -1;
          }
          else
          {
            LineFromHexTbToAsciiTb(CurrentLineChanged,cursor);
          }
        }



       // GetLine(richTextBoxHex, CurrentLineHex);
      }

    }

    private void checkHex(int cursor)
    {
      richTextBoxHex.Select(cursor, 1);
      /*Regex regHex = new Regex("^[0-9a-fA-F]$");
      if (richTextBoxHex.Text.Length > 0)
      {
        if ((!regHex.IsMatch(richTextBoxHex.Text)))
        {
          // richTextBoxHex.Text = richTextBoxHex.Text.Remove(richTextBoxHex.Text.Length - 1);
          richTextBoxHex.Select(cursor - 1, 1);
          richTextBoxHex.SelectedText = "";
          //richTextBoxHex.Focus();
          //richTextBoxHex.SelectionStart = cursor-1;
          return;
        }
      }*/
    }

    private string GetLine(RichTextBox Tb, int Line)
    {
      int start = Tb.GetFirstCharIndexFromLine(Line);
      int lengthToSelect;
      if (Line != nb_blocks-1)
      {
        lengthToSelect =
        Tb.GetFirstCharIndexFromLine(Line + 1)
      - Tb.GetFirstCharIndexFromLine(Line) - 1;
      }
      else
      {
        lengthToSelect = Tb.Text.Length
        - Tb.GetFirstCharIndexFromLine(Line) ;
      }
      Tb.Select(start, lengthToSelect);
      return (Tb.SelectedText);
      //MessageBox.Show(Tb.SelectedText);
    }

    private void RollBackLineHex(RichTextBox Tb, int Line)
    {
      int start = Tb.GetFirstCharIndexFromLine(Line);
      int lengthToSelect;
      if (Line != nb_blocks - 1)
      {
        lengthToSelect =
        richTextBoxHex.GetFirstCharIndexFromLine(Line + 1)
      - richTextBoxHex.GetFirstCharIndexFromLine(Line) - 1;
      }
      else
      {
        lengthToSelect = richTextBoxHex.Text.Length
        - richTextBoxHex.GetFirstCharIndexFromLine(Line);
      }
      Tb.Select(start, lengthToSelect);
      string newText = "";
      for (int i = 0; i < sz_blocks; i++)
      {
        newText +=
           String.Format("{0:X02}", cardContent[Line * sz_blocks + i]);
        if (i < sz_blocks - 1)
          newText += " ";
      }
      checkLine = false;
      Tb.SelectedText = newText;
      checkLine = true;

    }

    private void RollBackLineAscii(RichTextBox Tb, int Line)
    {
      int start = Tb.GetFirstCharIndexFromLine(Line);
      int lengthToSelect;
      if (Line != nb_blocks - 1)
      {
        lengthToSelect =
        richTextBoxASCII.GetFirstCharIndexFromLine(Line + 1)
      - richTextBoxASCII.GetFirstCharIndexFromLine(Line) - 1;
      }
      else
      {
        lengthToSelect = richTextBoxASCII.Text.Length
        - richTextBoxASCII.GetFirstCharIndexFromLine(Line);
      }
      Tb.Select(start, lengthToSelect);
      string newText = "";
      for (int i = 0; i < sz_blocks; i++)
        newText +=
           byteToAscii(cardContent[Line * sz_blocks + i]);
      checkLine = false;
      Tb.SelectedText = newText;
      checkLine = true;

    }

    private void LineFromHexTbToAsciiTb(int Line, int cursorPosition)
    {
      string TextHex = GetLine(richTextBoxHex, Line);
      string TextAscii = "";
      char[] delimiter = { ' ' };
      string[] DataHex = TextHex.Split(delimiter);

      for (int i = 0; i < sz_blocks; i++)
      {
        if (DataHex[i] != "")
          TextAscii += byteToAscii(
            System.Convert.ToByte(int.Parse(
            DataHex[i],
            System.Globalization.NumberStyles.HexNumber)));
        else
          TextAscii += "  ";
      }

      int lengthToSelect;
      if (Line != nb_blocks - 1)
      {
        lengthToSelect =
        richTextBoxASCII.GetFirstCharIndexFromLine(Line + 1)
      - richTextBoxASCII.GetFirstCharIndexFromLine(Line) - 1;
      }
      else
      {
        lengthToSelect = richTextBoxASCII.Text.Length
        - richTextBoxASCII.GetFirstCharIndexFromLine(Line);
      }

      richTextBoxASCII.Select(
        richTextBoxASCII.GetFirstCharIndexFromLine(Line),
        lengthToSelect);
      checkLine = false;
      richTextBoxASCII.SelectedText = TextAscii;
      checkLine = true;
      richTextBoxHex.DeselectAll();
      richTextBoxHex.SelectionStart = cursorPosition;
      
    }

    private void LineFromAsciiTbToHexTb(int Line, int cursorPosition)
    {
      string DataAscii = GetLine(richTextBoxASCII, Line);
      string TextHex = "";
      //char[] delimiter = { '' };
      //string[] DataAscii = TextAscii.Split(delimiter);

      for (int i = 0; i < sz_blocks; i++)
      {
        
        TextHex += String.Format("{0:X02}",
           (byte)
              System.Convert.ToChar(
              DataAscii[i])) + " ";
      }
      int lengthToSelect;
      if (Line != nb_blocks - 1)
      {
        lengthToSelect =
        richTextBoxHex.GetFirstCharIndexFromLine(Line + 1)
      - richTextBoxHex.GetFirstCharIndexFromLine(Line) - 1;
      }
      else
      {
         lengthToSelect = richTextBoxHex.Text.Length
         - richTextBoxHex.GetFirstCharIndexFromLine(Line);
      }
      richTextBoxHex.Select(
        richTextBoxHex.GetFirstCharIndexFromLine(Line),
        lengthToSelect);
      checkLine = false;
      richTextBoxHex.SelectedText = TextHex;
      checkLine = true;
      richTextBoxASCII.DeselectAll();
      richTextBoxASCII.SelectionStart = cursorPosition;

    }

    private void richTextBoxASCII_TextChanged(object sender, EventArgs e)
    {
      int cursor = richTextBoxASCII.SelectionStart;
      if (checkLine)
      {
        if ((CurrentLineChanged == -1)
          || (CurrentLineChanged == richTextBoxASCII.GetLineFromCharIndex(
            richTextBoxASCII.SelectionStart)))
        {
          CurrentLineChanged = richTextBoxASCII.GetLineFromCharIndex(
              richTextBoxASCII.SelectionStart);
          toolStripLabel.Text = "You are modifying line " + CurrentLineChanged.ToString();
          if (writable[
            richTextBoxASCII.GetLineFromCharIndex(
            richTextBoxASCII.SelectionStart)] == "red")
          {
            MessageBox.Show("Data from this line cannot be changed.");

            RollBackLineAscii(richTextBoxASCII, CurrentLineChanged);
            CurrentLineChanged = -1;
          }
          else
          {

            
            LineFromAsciiTbToHexTb(CurrentLineChanged, cursor);
          }
        }
      }
    }

    private void richTextBoxHex_Click(object sender, EventArgs e)
    {
      skipSpacesBack(richTextBoxHex,false);
    }

    private void write()
    {
      if (CurrentLineChanged == -1)
      {
        MessageBox.Show("Nothing to write");
        return;
      }
      
      if (writable[CurrentLineChanged] == "orange")
      {
        /* Ask before Writing */
        if (MessageBox.Show("It is not recommended that you write data on this block.\n" +
    "Make sure that you have fully understood card documentation before writing.\n" +
    "Continue anyway?", "Warning", MessageBoxButtons.OKCancel) != DialogResult.Cancel)
          WriteLine(CurrentLineChanged);
      }
      else
      {
        WriteLine(CurrentLineChanged);
      }
      toolStripLabel.Text = "Writing line " + CurrentLineChanged.ToString() +
        " returned : " + lastApduOut;
      CurrentLineChanged = -1;
      
    }


    private void skipSpacesBack(RichTextBox Tb, bool spaceAllowed)
    {
      int Currentcursor = Tb.SelectionStart;
      SelectOneChar(Tb);
      if (Tb.SelectedText == System.Convert.ToChar(0x0A).ToString())
        Tb.Select(Currentcursor - 1, 1);
      if (!spaceAllowed)
      {
        if (Tb.SelectedText == " ")
          Tb.Select(Currentcursor - 1, 1);
      }
      /*richTextBoxHex.DeselectAll();
      richTextBoxHex.SelectionStart = newCursor;*/
      
    }

    private void skipSpacesFw(RichTextBox Tb, bool spaceAllowed)
    {
      int Currentcursor = Tb.SelectionStart;
      SelectOneChar(Tb);
      if (Tb.SelectedText == System.Convert.ToChar(0x0A).ToString())
        Tb.Select(Currentcursor + 1, 1);
      if (!spaceAllowed)
      {
        if (Tb.SelectedText == " ")
          Tb.Select(Currentcursor + 1, 1);
      }
      /*richTextBoxHex.DeselectAll();
      richTextBoxHex.SelectionStart = newCursor;*/

    }

    private void richTextBoxHex_KeyDown(object sender, KeyEventArgs e)
    {
      int cursor = richTextBoxHex.SelectionStart;


      /* moving selection to the left */
      if (((e.KeyCode == Keys.Left)
        /*|| (e.KeyCode == Keys.Up)
        || (e.KeyCode == Keys.Down)*/) && (cursor>0))
       
      {
        richTextBoxHex.SelectionStart = cursor - 1;
        skipSpacesBack(richTextBoxHex,false);
        /*richTextBoxHex.DeselectAll();
        richTextBoxHex.Select(cursor-1 , 1);
        if (richTextBoxHex.SelectedText == " ")
          richTextBoxHex.Select(cursor - 2, 1);*/
      }
      else
      {
        if ((e.KeyCode != Keys.Right) && (e.KeyCode != Keys.Up)
        && (e.KeyCode != Keys.Down) &&
            (e.KeyCode != Keys.A) && (e.KeyCode != Keys.B) &&
            (e.KeyCode != Keys.C) && (e.KeyCode != Keys.D) &&
            (e.KeyCode != Keys.E) && (e.KeyCode != Keys.F) &&
            (e.KeyCode != Keys.D0) && (e.KeyCode != Keys.D1) &&
            (e.KeyCode != Keys.D2) && (e.KeyCode != Keys.D3) &&
            (e.KeyCode != Keys.D4) && (e.KeyCode != Keys.D5) &&
            (e.KeyCode != Keys.D6) && (e.KeyCode != Keys.D7) &&
            (e.KeyCode != Keys.D8) && (e.KeyCode != Keys.D9) &&
            (e.KeyCode != Keys.NumPad0) && (e.KeyCode != Keys.NumPad1) && 
            (e.KeyCode != Keys.NumPad2) && (e.KeyCode != Keys.NumPad3) &&  
            (e.KeyCode != Keys.NumPad4) && (e.KeyCode != Keys.NumPad5) && 
            (e.KeyCode != Keys.NumPad6) && (e.KeyCode != Keys.NumPad7) && 
            (e.KeyCode != Keys.NumPad8) && (e.KeyCode != Keys.NumPad9)  
          )
          e.SuppressKeyPress = true;
      }



    }

    private void richTextBoxHex_KeyUp(object sender, KeyEventArgs e)
    {
      if ((e.KeyCode == Keys.Down) || (e.KeyCode == Keys.Up))
        richTextBoxHex.SelectionStart--;
      skipSpacesFw(richTextBoxHex,false);
      int cursor = richTextBoxHex.SelectionStart;
      if ((checkLine) && (CurrentLineChanged != -1) &&
       (CurrentLineChanged != richTextBoxHex.GetLineFromCharIndex(
           richTextBoxHex.SelectionStart))) 
        {
          String message =
          "You haven't recorded line " +
          CurrentLineChanged.ToString() + "\n" + "Would you like to write it to card ?";

          if (MessageBox.Show(message, "Record", MessageBoxButtons.YesNo)
            == DialogResult.No)
          {
            /* Rollback Changes made to CurrentLine*/
            RollBackLineHex(richTextBoxHex, CurrentLineChanged);
            RollBackLineAscii(richTextBoxASCII, CurrentLineChanged);
          }
          else
          {
            write();
          }
          CurrentLineChanged = -1;
          richTextBoxHex.SelectionStart = cursor;
          SelectOneChar(richTextBoxHex);
        }
    }

    private void richTextBoxASCII_Click(object sender, EventArgs e)
    {
      //SelectOneChar(richTextBoxASCII);
      skipSpacesBack(richTextBoxASCII, true);
    }

    private void SelectOneChar(RichTextBox Tb)
    {
      int Currentcursor = Tb.SelectionStart;
      if (Currentcursor == Tb.Text.Length)
        Currentcursor--;
      int newCursor = Currentcursor;
      Tb.DeselectAll();
      Tb.Select(Currentcursor, 1);
    }

    private void richTextBoxASCII_KeyUp(object sender, KeyEventArgs e)
    {
      
      //SelectOneChar(richTextBoxASCII);
      if ((e.KeyCode == Keys.Down) || (e.KeyCode == Keys.Up))
        richTextBoxASCII.SelectionStart--;
      skipSpacesFw(richTextBoxASCII,true);
      int cursor = richTextBoxASCII.SelectionStart;
      if ((checkLine) && (CurrentLineChanged != -1) &&
        (CurrentLineChanged != richTextBoxASCII.GetLineFromCharIndex(
            richTextBoxASCII.SelectionStart)))
      {
        String message =
        "You haven't recorded line " +
        CurrentLineChanged.ToString() + "\n" + "Would you like to write it to card ?";

        if (MessageBox.Show(message, "Record", MessageBoxButtons.YesNo)
          == DialogResult.No)
        {
          RollBackLineHex(richTextBoxHex, CurrentLineChanged);
          RollBackLineAscii(richTextBoxASCII, CurrentLineChanged);
        }
        else
        {
          write();
        }
        CurrentLineChanged = -1;
        richTextBoxASCII.SelectionStart = cursor;
        SelectOneChar(richTextBoxASCII);
      }
    }

    private void richTextBoxASCII_KeyDown(object sender, KeyEventArgs e)
    {
      int cursor = richTextBoxASCII.SelectionStart;

      if ((e.KeyCode == Keys.Back) || (e.KeyCode == Keys.Delete))
        e.SuppressKeyPress = true;
      if (((e.KeyCode == Keys.Left)
      // || (e.KeyCode == Keys.Up)
       /*|| (e.KeyCode == Keys.Down)*/ )&& (cursor>0))
      {
        richTextBoxASCII.SelectionStart = cursor - 1;
        skipSpacesBack(richTextBoxASCII,true);
      }

    }
#endregion

#region card status
    /* check every 100ms card status for selected reader, 
     * and change interface according to status found */

    System.Threading.Thread StatusThread = null; /* Get reader status */
    public StatusThreadFinishedDelegate ThStatusDeleg;
    public delegate void StatusThreadFinishedDelegate();
    public string StatusStr = "";

    public void ThStatusProc()
    {
      string str = SPCSC.CheckStatus();
      StatusStr = str;
      if (str.Contains(":"))
        StatusStr = StatusStr.Remove(str.IndexOf(':'));
      try
      {
        Invoke(ThStatusDeleg);
      }
      catch (Exception e) { }

    }

    private void timerCard_Tick(object sender, EventArgs e)
    {
      StatusThread = new System.Threading.Thread(new
      System.Threading.ThreadStart(ThStatusProc));
      StatusThread.Start();
    }

    public void onStatusDone()
    {
     string status = StatusStr;
      if (status == "")
      {
        CardRemoved();
      }

      if (status.Contains("EMPTY"))
      {
        CardRemoved();
      }
    }

    public void CardRemoved()
    {
      timerCard.Enabled = false;
      MessageBox.Show("Card has been removed.\nRead/Write operations disabled.");
      //ButStart.Enabled = false;
      //SetKeysBut.Enabled = false;
      richTextBoxHex.ReadOnly = true;
      richTextBoxASCII.ReadOnly = true;
      toolStripLabel.Text = "CARD HAS BEEN REMOVED. READ/WRITE OPERATIONS DISABLED.";
      
    }

#endregion

#region menu bar

    private void aboutToolStripMenuItem_Click(object sender, EventArgs e)
    {
      AboutBoxFrm a = new AboutBoxFrm();
      a.Show();
    }


    private void readToolStripMenuItem_Click(object sender, EventArgs e)
    {
      StartReading();
    }


    private void pauseToolStripMenuItem_Click(object sender, EventArgs e)
    {
      if (ReadThread != null)
      {
        try
        {
          ReadThread.Suspend();
          SPCSC.Disconnect();
          FullLoadDataTable();
        }
        catch (Exception x)
        {
          MessageBox.Show(x.Message);
        }
      }
    }


    private void setKeysToolStripMenuItem_Click(object sender, EventArgs e)
    {
      KeysDlg = new KeysFrm(this, keys);
      KeysDlg.Show();
    }

    private void quitToolStripMenuItem_Click(object sender, EventArgs e)
    {
      this.Close();
      this.Dispose();
    }

    private void writeToolStripMenuItem_Click(object sender, EventArgs e)
    {
      write();
    }

#endregion

#region resize

    private void panelMain_Resize(object sender, EventArgs e)
    {
      int newWidth = panelMain.Size.Width;
      int widthAscii = ciAsciiWidth * sz_blocks;
      if (widthAscii < minAscii)
        widthAscii = minAscii;
      int widthHex = ciHexWidth * sz_blocks;
      if (newWidth < 792)
      /*we have to resize !*/
      {
        int delta = (792 - newWidth) / 3;
       // MessageBox.Show(delta.ToString());
        int nwidthHex = widthHex - delta;
        if (nwidthHex < 50)
          nwidthHex = 50;

        richTextBoxHex.Size = new Size(
          nwidthHex,
          richTextBoxHex.Size.Height);
        int nwidthAscii = widthAscii - delta;
        if (nwidthAscii < 50)
          nwidthAscii = 50;

        richTextBoxASCII.Location = new Point(
          richTextBoxHex.Location.X + richTextBoxHex.Size.Width + 5,
          richTextBoxASCII.Location.Y);
        richTextBoxASCII.Size = new Size(
          nwidthAscii,
          richTextBoxASCII.Size.Height);
        richTextBoxInfo.Location = new Point(
          richTextBoxASCII.Location.X + richTextBoxASCII.Size.Width + 5,
          richTextBoxInfo.Location.Y);

        int nwidthInfo = richTextBoxInfo.Size.Width - delta;
        if (nwidthInfo < 50)
          nwidthInfo = 50;
        richTextBoxInfo.Size = new Size(
           nwidthInfo,
          richTextBoxInfo.Size.Height);
      }


    }
#endregion


    private void RWCardFrm_FormClosing(object sender, FormClosingEventArgs e)
    {
      parentFrm.Restart_Timer();
    }









  }
}