using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Runtime.InteropServices;
using SpringPCSCAPI;

/*
 * In this form we consider that the card is a Mifare Ultralight C
 * 
 * If not then it is a plain Mifare Ultralight 
 *
 */

namespace CardTool
{
  public partial class RWMifUlFrm : Form
  {
    DataTable CardDataTable;
    DataTable CardAsciiTable;
    int nb_blocks = 44;
    int sz_blocks = 4;
    int hCard;
    byte[] cardContent;
    string InfoText = "";
    int CurrentLine = -1;
    int CurrentLineChanged = -1;
    string[] writable;
    bool checkLine = false;
    ReaderStatusFrm parentFrm;
    private const float ciLineHeight = 14.32F;

    public byte[] BLANK_KEY =
new byte[16] { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
      0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };

    public byte[] TRANSPORT_KEY =
new byte[16] { 0x49, 0x45, 0x4D, 0x4B, 0x41, 0x45, 0x52, 0x42,
                             0x21, 0x4E, 0x41, 0x43, 0x55, 0x4F, 0x59, 0x46 };

    public RWMifUlFrm(ReaderStatusFrm parent)
    {
      InitializeComponent();
      parentFrm = parent;
      cardContent = new byte[sz_blocks * nb_blocks];
      /* mousewheel */
      richTextBoxASCII.MouseWheel += new MouseEventHandler(richTextBoxASCII_MouseWheel);
      richTextBoxHex.MouseWheel += new MouseEventHandler(richTextBoxHex_MouseWheel);
      richTextBoxBlock.MouseWheel += new MouseEventHandler(richTextBoxBlock_MouseWheel);
      richTextBoxInfo.MouseWheel += new MouseEventHandler(richTextBoxInfo_MouseWheel);

      ThStatusDeleg = new StatusThreadFinishedDelegate(onStatusDone);

      /* writable array contains indexes of writable blocks for mifare Ultralight
       * using color codes green (writable), orange (write with care) and red 
       * (cannot be written)
       */
      writable = new string[nb_blocks];
      writable[0] = "red";
      writable[1] = "red";
      writable[2] = "red";
      writable[3] = "orange";
      for (int i = 4; i < 40; i++)
        writable[i] = "green";
      for (int i = 40; i < nb_blocks; i++)
        writable[i] = "orange";
      SPCSC.ConnectCard();
      hCard = SPCSC.GetHCard();
      AttachLibrary(hCard);

      StartReading();
    }


    private void StartReading()
    {
      checkLine = false;
      MifUlC_ReadFullCard();
      FullLoadDataTable();
      checkLine = true;
    }


#region MouseWheel
    void richTextBoxInfo_MouseWheel(object sender, MouseEventArgs e)
    {
      panelMain.Focus();
    }

    void richTextBoxBlock_MouseWheel(object sender, MouseEventArgs e)
    {
      panelMain.Focus();
    }

    void richTextBoxHex_MouseWheel(object sender, MouseEventArgs e)
    {
      panelMain.Focus();
    }

    void richTextBoxASCII_MouseWheel(object sender, MouseEventArgs e)
    {
      panelMain.Focus();
    }
#endregion

#region DLL Imports
    // Attach Library : use this function before using any function of the DLL
    [DllImport("pcsc_mifulc.dll", EntryPoint = "SCardMifUlC_AttachLibrary", SetLastError = true)]
    internal static extern uint AttachLibrary(int hCard);

    // Detach Library : use this function after using functions of the DLL
    [DllImport("pcsc_mifulc.dll", EntryPoint = "SCardMifUlC_DetachLibrary", SetLastError = true)]
    internal static extern uint DetachLibrary(int hCard);

    // Authenticate : use this function to authenticate with a 16-byte key
    [DllImport("pcsc_mifulc.dll", EntryPoint = "SCardMifUlC_Authenticate", SetLastError = true)]
    internal static extern uint Authenticate(int hCard, byte[] key);

    // Read : read 4 blocks in a row (16 bytes)
    [DllImport("pcsc_mifulc.dll", EntryPoint = "SCardMifUlC_Read", SetLastError = true)]
    internal static extern uint Read(int hCard,byte page,[In,Out] byte[] data);

    // Write : write 4 blocks in a row (16 bytes)
    [DllImport("pcsc_mifulc.dll", EntryPoint = "SCardMifUlC_Write", SetLastError = true)]
    internal static extern uint Write(int hCard, byte page, byte[] data);

    // Write4 : write one block (4 bytes)
    [DllImport("pcsc_mifulc.dll", EntryPoint = "SCardMifUlC_Write4", SetLastError = true)]
    internal static extern uint Write4(int hCard, byte page, byte[] data);



#endregion

#region Mifare UltralightC functions : Authenticate, Read, Write
    private void MifUlC_Authenticate()
    {
      // let's try to authenticate with our keys
      uint  rc = Authenticate(hCard, BLANK_KEY);
      if (rc != 0)
      {
        InfoText += "Cannot Authenticate with blank_key \n";

        rc = Authenticate(hCard, TRANSPORT_KEY);
        if (rc != 0)
          InfoText += "Cannot Authenticate with transport_key\n";
        else
          InfoText += "Authenticate with transport_key successfull\n";



      }
      else
        InfoText += "Authenticate with blank_key successfull \n";

    
    }

    private void MifUlC_ReadFullCard()
    {
      byte page;
      byte[] data = new byte[16];
      uint rc;
      int offset_cardContent = 0;
      byte[] tab_temp = new byte[cardContent.Length];
      for (page = 0; page < 44; page += 4)
      {
        MifUlC_Authenticate();
        rc = Read(hCard, page, data);

        if (rc != 0)
        {
          InfoText += "Error Reading Page"+page.ToString()+"\n";
          nb_blocks = page;
          break;
        }
        for (int i = 0; i < 16; i++)
        {
          if(offset_cardContent+i<cardContent.Length)
           tab_temp[offset_cardContent + i] = data[i];
          
        }
        offset_cardContent += 16;
      }
      cardContent = new byte[offset_cardContent];
      for (int cpt = 0; cpt < offset_cardContent; cpt++)
        cardContent[cpt] = tab_temp[cpt];
    }

    // write a block (4 bytes)
    private void MifUlC_WriteBlock(byte index, byte[] block_content)
    {
      uint rc;
      if (block_content.Length == 4)
      {
        MifUlC_Authenticate();
        rc = Write4(hCard, index, block_content);
        if (rc != 0)
          InfoText = "Error Writing Line " + index.ToString() +
            " : "+rc.ToString()+"\n";
        else
          InfoText = "Line " + index.ToString() + " successfully written\n";

      }
      
    }

#endregion

    private void write()
    {
      if (CurrentLineChanged == -1)
      {
        MessageBox.Show("Nothing to write");
        return;
      }

      if (writable[CurrentLineChanged] == "orange")
      {
        /* Ask before Writing */
        if (MessageBox.Show("It is not recommended that you write data on this block.\n" +
    "Make sure that you have fully understood card documentation before writing.\n" +
    "Continue anyway?", "Warning", MessageBoxButtons.OKCancel) != DialogResult.Cancel)
          WriteLine(CurrentLineChanged);
      }
      else
      {
        WriteLine(CurrentLineChanged);
      }

      CurrentLineChanged = -1;

    }


    public void WriteLine(int line)
    {
      // NewData is the data to be written on the card
      byte[] NewData = new byte[sz_blocks];
      
      char[] NewLine = { '\n' };
      string[] Line = GetLine(richTextBoxHex, line).Split(NewLine);
      string TextHex = Line[0];
      string TextAscii = "";
      char[] delimiter = { ' ' };
      string[] DataHex = TextHex.Split(delimiter);

      // retreive data from datagrid
      for (int i = 0; i < sz_blocks; i++)
        NewData[i] =
          System.Convert.ToByte(int.Parse(
          DataHex[i],
          System.Globalization.NumberStyles.HexNumber));
      MifUlC_WriteBlock((byte) line, NewData);
      richTextBoxInfo.Text += InfoText;
      // scroll bar
      // Put the cursor at the very end of the text
      richTextBoxInfo.Select((richTextBoxInfo.Text.Length), 0);

      // Scroll
      richTextBoxInfo.ScrollToCaret();

      CurrentLine = -1;
      // change CardContent Array & CardContentAscii
      for (int i = 0; i < sz_blocks; i++)
      {
        cardContent[line * sz_blocks + i] = NewData[i];
      }
    }


#region RichTextBoxes management
    public void FullLoadDataTable()
    {
      int i, j, cpt = 0;


      CardDataTable = new DataTable();
      CardAsciiTable = new DataTable();

      int height = (int)(ciLineHeight * nb_blocks);


      richTextBoxASCII.Size = new Size(richTextBoxASCII.Size.Width,
  height);
      richTextBoxHex.Size = new Size(richTextBoxHex.Size.Width,
       height);
      richTextBoxBlock.Size = new Size(richTextBoxBlock.Size.Width,
        height);
      richTextBoxInfo.Size = new Size(richTextBoxInfo.Size.Width,
        height);

      richTextBoxBlock.Clear();
      richTextBoxASCII.Clear();
      richTextBoxHex.Clear();
      for (i = 0; i < nb_blocks; i++)
      {

        richTextBoxBlock.AppendText(i.ToString());
        for (j = 0; j < sz_blocks; j++)
        {
          if (cpt < cardContent.Length)
          {
            richTextBoxHex.AppendText(String.Format("{0:X02}", cardContent[cpt]));
            if (j < sz_blocks - 1)
              richTextBoxHex.AppendText(" ");

            richTextBoxASCII.AppendText(byteToAscii(cardContent[cpt]));

            cpt++;
          }
          else
            break;
        }

        if (i < nb_blocks - 1)
        {
          richTextBoxBlock.AppendText("\n");
          richTextBoxASCII.AppendText("\n");
          richTextBoxHex.AppendText("\n");
        }
      }

      richTextBoxInfo.Text = InfoText;
      checkLine = true;

      panelMain.Focus();
    }

    private void richTextBoxHex_TextChanged(object sender, EventArgs e)
    {

      int cursor = richTextBoxHex.SelectionStart;
      if (checkLine)
      {


        if ((CurrentLineChanged == -1)
          || (CurrentLineChanged == richTextBoxHex.GetLineFromCharIndex(
            richTextBoxHex.SelectionStart)))
        {
          CurrentLineChanged = richTextBoxHex.GetLineFromCharIndex(
              richTextBoxHex.SelectionStart);
          toolStripLabel.Text = "You are modifying line " + CurrentLineChanged.ToString();
          if (writable[
            richTextBoxHex.GetLineFromCharIndex(
            richTextBoxHex.SelectionStart)] == "red")
          {
            MessageBox.Show("Data from this line cannot be changed.");

            RollBackLineHex(richTextBoxHex, CurrentLineChanged);
            CurrentLineChanged = -1;
          }
          else
          {
            LineFromHexTbToAsciiTb(CurrentLineChanged, cursor);
          }
        }


      }

    }

    private void checkHex(int cursor)
    {
      richTextBoxHex.Select(cursor, 1);
    }

    private string GetLine(RichTextBox Tb, int Line)
    {
      int start = Tb.GetFirstCharIndexFromLine(Line);
      int lengthToSelect;
      if (Line != nb_blocks - 1)
      {
        lengthToSelect =
        Tb.GetFirstCharIndexFromLine(Line + 1)
      - Tb.GetFirstCharIndexFromLine(Line) - 1;
      }
      else
      {
        lengthToSelect = Tb.Text.Length
        - Tb.GetFirstCharIndexFromLine(Line);
      }
      Tb.Select(start, lengthToSelect);
      return (Tb.SelectedText);
    }

    private void RollBackLineHex(RichTextBox Tb, int Line)
    {
      int start = Tb.GetFirstCharIndexFromLine(Line);
      int lengthToSelect;
      if (Line != nb_blocks - 1)
      {
        lengthToSelect =
        richTextBoxHex.GetFirstCharIndexFromLine(Line + 1)
      - richTextBoxHex.GetFirstCharIndexFromLine(Line) - 1;
      }
      else
      {
        lengthToSelect = richTextBoxHex.Text.Length
        - richTextBoxHex.GetFirstCharIndexFromLine(Line);
      }
      Tb.Select(start, lengthToSelect);
      string newText = "";
      for (int i = 0; i < sz_blocks; i++)
      {
        newText +=
           String.Format("{0:X02}", cardContent[Line * sz_blocks + i]);
        if (i < sz_blocks - 1)
          newText += " ";
      }
      checkLine = false;
      Tb.SelectedText = newText;
      checkLine = true;

    }

    private void RollBackLineAscii(RichTextBox Tb, int Line)
    {
      int start = Tb.GetFirstCharIndexFromLine(Line);
      int lengthToSelect;
      if (Line != nb_blocks - 1)
      {
        lengthToSelect =
        richTextBoxASCII.GetFirstCharIndexFromLine(Line + 1)
      - richTextBoxASCII.GetFirstCharIndexFromLine(Line) - 1;
      }
      else
      {
        lengthToSelect = richTextBoxASCII.Text.Length
        - richTextBoxASCII.GetFirstCharIndexFromLine(Line);
      }
      Tb.Select(start, lengthToSelect);
      string newText = "";
      for (int i = 0; i < sz_blocks; i++)
        newText +=
           byteToAscii(cardContent[Line * sz_blocks + i]);
      checkLine = false;
      Tb.SelectedText = newText;
      checkLine = true;

    }

    private void LineFromHexTbToAsciiTb(int Line, int cursorPosition)
    {
      string TextHex = GetLine(richTextBoxHex, Line);
      string TextAscii = "";
      char[] delimiter = { ' ' };
      string[] DataHex = TextHex.Split(delimiter);

      for (int i = 0; i < sz_blocks; i++)
      {
        if (DataHex[i] != "")
          TextAscii += byteToAscii(
            System.Convert.ToByte(int.Parse(
            DataHex[i],
            System.Globalization.NumberStyles.HexNumber)));
        else
          TextAscii += "  ";
      }

      int lengthToSelect;
      if (Line != nb_blocks - 1)
      {
        lengthToSelect =
        richTextBoxASCII.GetFirstCharIndexFromLine(Line + 1)
      - richTextBoxASCII.GetFirstCharIndexFromLine(Line) - 1;
      }
      else
      {
        lengthToSelect = richTextBoxASCII.Text.Length
        - richTextBoxASCII.GetFirstCharIndexFromLine(Line);
      }

      richTextBoxASCII.Select(
        richTextBoxASCII.GetFirstCharIndexFromLine(Line),
        lengthToSelect);
      checkLine = false;
      richTextBoxASCII.SelectedText = TextAscii;
      checkLine = true;
      richTextBoxHex.DeselectAll();
      richTextBoxHex.SelectionStart = cursorPosition;

    }

    private void LineFromAsciiTbToHexTb(int Line, int cursorPosition)
    {
      string DataAscii = GetLine(richTextBoxASCII, Line);
      string TextHex = "";

      for (int i = 0; i < sz_blocks; i++)
      {

        TextHex += String.Format("{0:X02}",
           (byte)
              System.Convert.ToChar(
              DataAscii[i])) + " ";
      }
      int lengthToSelect;
      if (Line != nb_blocks - 1)
      {
        lengthToSelect =
        richTextBoxHex.GetFirstCharIndexFromLine(Line + 1)
      - richTextBoxHex.GetFirstCharIndexFromLine(Line) - 1;
      }
      else
      {
        lengthToSelect = richTextBoxHex.Text.Length
        - richTextBoxHex.GetFirstCharIndexFromLine(Line);
      }
      richTextBoxHex.Select(
        richTextBoxHex.GetFirstCharIndexFromLine(Line),
        lengthToSelect);
      checkLine = false;
      richTextBoxHex.SelectedText = TextHex;
      checkLine = true;
      richTextBoxASCII.DeselectAll();
      richTextBoxASCII.SelectionStart = cursorPosition;

    }

    private void richTextBoxASCII_TextChanged(object sender, EventArgs e)
    {
      int cursor = richTextBoxASCII.SelectionStart;
      if (checkLine)
      {
        if ((CurrentLineChanged == -1)
          || (CurrentLineChanged == richTextBoxASCII.GetLineFromCharIndex(
            richTextBoxASCII.SelectionStart)))
        {
          CurrentLineChanged = richTextBoxASCII.GetLineFromCharIndex(
              richTextBoxASCII.SelectionStart);
          toolStripLabel.Text = "You are modifying line " + CurrentLineChanged.ToString();
          if (writable[
            richTextBoxASCII.GetLineFromCharIndex(
            richTextBoxASCII.SelectionStart)] == "red")
          {
            MessageBox.Show("Data from this line cannot be changed.");

            RollBackLineAscii(richTextBoxASCII, CurrentLineChanged);
            CurrentLineChanged = -1;
          }
          else
          {


            LineFromAsciiTbToHexTb(CurrentLineChanged, cursor);
          }
        }

      }
    }

    private void richTextBoxHex_Click(object sender, EventArgs e)
    {
      skipSpacesBack(richTextBoxHex, false);
    }

    private void skipSpacesBack(RichTextBox Tb, bool spaceAllowed)
    {
      int Currentcursor = Tb.SelectionStart;
      SelectOneChar(Tb);
      if (Tb.SelectedText == System.Convert.ToChar(0x0A).ToString())
        Tb.Select(Currentcursor - 1, 1);
      if (!spaceAllowed)
      {
        if (Tb.SelectedText == " ")
          Tb.Select(Currentcursor - 1, 1);
      }

    }

    private void skipSpacesFw(RichTextBox Tb, bool spaceAllowed)
    {
      int Currentcursor = Tb.SelectionStart;
      SelectOneChar(Tb);
      if (Tb.SelectedText == System.Convert.ToChar(0x0A).ToString())
        Tb.Select(Currentcursor + 1, 1);
      if (!spaceAllowed)
      {
        if (Tb.SelectedText == " ")
          Tb.Select(Currentcursor + 1, 1);
      }
    }

    private void richTextBoxHex_KeyDown(object sender, KeyEventArgs e)
    {
      int cursor = richTextBoxHex.SelectionStart;


      /* moving selection to the left */
      if ((e.KeyCode == Keys.Left)&& (cursor > 0))
      {
        richTextBoxHex.SelectionStart = cursor - 1;
        skipSpacesBack(richTextBoxHex, false);
      }
      else
      {
        if ((e.KeyCode != Keys.Right) && (e.KeyCode != Keys.Up)
        && (e.KeyCode != Keys.Down) &&
            (e.KeyCode != Keys.A) && (e.KeyCode != Keys.B) &&
            (e.KeyCode != Keys.C) && (e.KeyCode != Keys.D) &&
            (e.KeyCode != Keys.E) && (e.KeyCode != Keys.F) &&
            (e.KeyCode != Keys.D0) && (e.KeyCode != Keys.D1) &&
            (e.KeyCode != Keys.D2) && (e.KeyCode != Keys.D3) &&
            (e.KeyCode != Keys.D4) && (e.KeyCode != Keys.D5) &&
            (e.KeyCode != Keys.D6) && (e.KeyCode != Keys.D7) &&
            (e.KeyCode != Keys.D8) && (e.KeyCode != Keys.D9) &&
            (e.KeyCode != Keys.NumPad0) && (e.KeyCode != Keys.NumPad1) &&
            (e.KeyCode != Keys.NumPad2) && (e.KeyCode != Keys.NumPad3) &&
            (e.KeyCode != Keys.NumPad4) && (e.KeyCode != Keys.NumPad5) &&
            (e.KeyCode != Keys.NumPad6) && (e.KeyCode != Keys.NumPad7) &&
            (e.KeyCode != Keys.NumPad8) && (e.KeyCode != Keys.NumPad9)
          )
          e.SuppressKeyPress = true;
      }



    }

    private void richTextBoxHex_KeyUp(object sender, KeyEventArgs e)
    {
      if ((e.KeyCode == Keys.Down) || (e.KeyCode == Keys.Up))
        richTextBoxHex.SelectionStart--;
      skipSpacesFw(richTextBoxHex, false);
      int cursor = richTextBoxHex.SelectionStart;
      if ((checkLine) && (CurrentLineChanged != -1) &&
       (CurrentLineChanged != richTextBoxHex.GetLineFromCharIndex(
           richTextBoxHex.SelectionStart)))
      {
        String message =
        "You haven't recorded line " +
        CurrentLineChanged.ToString() + "\n" + "Would you like to write it to card ?";

        if (MessageBox.Show(message, "Record", MessageBoxButtons.YesNo)
          == DialogResult.No)
        {
          /* Rollback Changes made to CurrentLine*/
          RollBackLineHex(richTextBoxHex, CurrentLineChanged);
          RollBackLineAscii(richTextBoxASCII, CurrentLineChanged);
        }
        else
        {
          write();
        }
        CurrentLineChanged = -1;
        richTextBoxHex.SelectionStart = cursor;
        SelectOneChar(richTextBoxHex);
      }
    }

    private void richTextBoxASCII_Click(object sender, EventArgs e)
    {
      //SelectOneChar(richTextBoxASCII);
      skipSpacesBack(richTextBoxASCII, true);
    }

    private void SelectOneChar(RichTextBox Tb)
    {
      int Currentcursor = Tb.SelectionStart;
      if (Currentcursor == Tb.Text.Length)
        Currentcursor--;
      int newCursor = Currentcursor;
      Tb.DeselectAll();
      Tb.Select(Currentcursor, 1);
    }

    private void richTextBoxASCII_KeyUp(object sender, KeyEventArgs e)
    {

      //SelectOneChar(richTextBoxASCII);
      if ((e.KeyCode == Keys.Down) || (e.KeyCode == Keys.Up))
        richTextBoxASCII.SelectionStart--;
      skipSpacesFw(richTextBoxASCII, true);
      int cursor = richTextBoxASCII.SelectionStart;
      if ((checkLine) && (CurrentLineChanged != -1) &&
        (CurrentLineChanged != richTextBoxASCII.GetLineFromCharIndex(
            richTextBoxASCII.SelectionStart)))
      {
        String message =
        "You haven't recorded line " +
        CurrentLineChanged.ToString() + "\n" + "Would you like to write it to card ?";

        if (MessageBox.Show(message, "Record", MessageBoxButtons.YesNo)
          == DialogResult.No)
        {
          RollBackLineHex(richTextBoxHex, CurrentLineChanged);
          RollBackLineAscii(richTextBoxASCII, CurrentLineChanged);
        }
        else
        {
          write();
        }
        CurrentLineChanged = -1;
        richTextBoxASCII.SelectionStart = cursor;
        SelectOneChar(richTextBoxASCII);
      }
    }

    private void richTextBoxASCII_KeyDown(object sender, KeyEventArgs e)
    {
      int cursor = richTextBoxASCII.SelectionStart;

      if ((e.KeyCode == Keys.Back) || (e.KeyCode == Keys.Delete))
        e.SuppressKeyPress = true;
      if (((e.KeyCode == Keys.Left)
        // || (e.KeyCode == Keys.Up)
        /*|| (e.KeyCode == Keys.Down)*/ ) && (cursor > 0))
      {
        richTextBoxASCII.SelectionStart = cursor - 1;
        skipSpacesBack(richTextBoxASCII, true);
      }
    }
#endregion

#region conversion from str to byte and vice versa

    private byte[] createByteArrayFromStr(string StrIn)
    {
      string Str = StrIn.Replace("0x", "").Replace(" ", "");

      if (!Str.Contains(" "))
      {
        string nStr = "";
        for (int cpt = 0; cpt < Str.Length; cpt = cpt + 2)
        {
          nStr = nStr + Str.Substring(cpt, 2) + " ";
        }
        Str = nStr;
      }

      // command stored into byte array
      char[] delimiterByte = new char[1];
      delimiterByte[0] = ' ';
      string properStr = Str;
      // removing last " "
      if (Str.EndsWith(" "))
      {
        properStr = Str.Substring(0, Str.Length - 1);
      }

      string[] Strbuff = properStr.Split(delimiterByte);
      byte[] StrBuffer = new byte[Strbuff.Length];
      for (int i = 0; i < Strbuff.Length; i++)
      {
        // converting byte string to byte  
        StrBuffer[i] = System.Convert.ToByte(int.Parse(Strbuff[i],
          System.Globalization.NumberStyles.HexNumber));
      }
      return StrBuffer;
    }

    private string createStrFromByteArray(byte[] byteArray)
    {
      string Str = "";
      if (byteArray.Length <= 0)
        return Str;
      for (int cpt = 0; cpt < byteArray.Length - 1; cpt++)
        Str += String.Format("{0:X02}", byteArray[cpt]) + " ";
      Str += String.Format("{0:X02}", byteArray[byteArray.Length - 1]);
      return Str;
    }

    private byte[] createByteArrayFromAsciiStr(string StrIn)
    {
      byte[] array = new byte[StrIn.Length];

      for (int cpt = 0; cpt < StrIn.Length; cpt++)
        array[cpt] = (byte)System.Convert.ToChar(StrIn[cpt]);

      return array;

    }

    private string createAsciiStrFromByteArray(byte[] byteArray)
    {
      string Str = "";
      if (byteArray.Length <= 0)
        return Str;
      for (int cpt = 0; cpt < byteArray.Length; cpt++)
        Str += byteToAscii(byteArray[cpt]);
      return Str;
    }

    public string byteToAscii(byte value)
    {
      if (value < 32 || value > 127)
        return (".");
      else
        return (System.Convert.ToChar(value).ToString());
    }

    public string byteToString(byte value)
    {
      if (value < 32 || value > 127)
        return (String.Format("{0:X02}", value));
      else
        return (System.Convert.ToChar(value).ToString());
    }

    public string byteToStringHex(byte value)
    {
      return ("0x" + String.Format("{0:X02}", value));
    }
#endregion

#region menu bar

    private void aboutToolStripMenuItem_Click(object sender, EventArgs e)
    {
      AboutBoxFrm a = new AboutBoxFrm();
      a.Show();
    }


    private void readToolStripMenuItem_Click(object sender, EventArgs e)
    {
      StartReading();
    }


     private void quitToolStripMenuItem_Click(object sender, EventArgs e)
    {
      this.Close();
      this.Dispose();
    }

    private void writeToolStripMenuItem_Click(object sender, EventArgs e)
    {
      write();
    }

#endregion

#region card status
    /* check every 100ms card status for selected reader, 
     * and change interface according to status found */

    System.Threading.Thread StatusThread = null; /* Get reader status */
    public StatusThreadFinishedDelegate ThStatusDeleg;
    public delegate void StatusThreadFinishedDelegate();
    public string StatusStr = "";

    public void ThStatusProc()
    {
      string str = SPCSC.CheckStatus();
      StatusStr = str;
      if (str.Contains(":"))
        StatusStr = StatusStr.Remove(str.IndexOf(':'));
      try
      {
        Invoke(ThStatusDeleg);
      }
      catch (Exception e) { }

    }

    private void timerCard_Tick(object sender, EventArgs e)
    {
      StatusThread = new System.Threading.Thread(new
      System.Threading.ThreadStart(ThStatusProc));
      StatusThread.Start();
    }

    public void onStatusDone()
    {
      string status = StatusStr;
      if (status == "")
      {
        CardRemoved();
      }

      if (status.Contains("EMPTY"))
      {
        CardRemoved();
      }
    }

    public void CardRemoved()
    {
      timerCard.Enabled = false;
      MessageBox.Show("Card has been removed.\nRead/Write operations disabled.");
      //ButStart.Enabled = false;
      //SetKeysBut.Enabled = false;
      richTextBoxHex.ReadOnly = true;
      richTextBoxASCII.ReadOnly = true;
      toolStripLabel.Text = "CARD HAS BEEN REMOVED. READ/WRITE OPERATIONS DISABLED.";

    }

#endregion

    private void RWMifUlFrm_FormClosing(object sender, FormClosingEventArgs e)
    {
      DetachLibrary(hCard);
      parentFrm.Restart_Timer();
    }

  }
}